/*
   See pages 176-179 from "Principles of Parallel Programming".
*/
#include <stdio.h>
#include <stdlib.h>

#ifdef MAC
   #include <OpenCL/cl.h>
#else
   #include <CL/cl.h>
#endif

#include "simpleCL.h"  // for sclPrintErrorFlag() function

// File name and function name for the OpenCL kernel
#define PROGRAM_FILE "2D-SOR-opencl-iterations.cl"
#define KERNEL_FUNC "two_d_sor"

#include "2D_SOR_float.h"
#include "initial_data_1_float.h"

/*
  Command line args: matrix size, iterations, PLATFORM_NUMBER
*/
int main(int argc, char*argv[])
{
   /* Host/device data structures */
   cl_platform_id platform[2]; // space for two platform data structures
   cl_device_id device;
   cl_context context;
   cl_command_queue cmdQueue;

   /* Program/kernel data structures */
   cl_program program;
   FILE *program_handle;
   char *program_buffer;
   size_t program_size;
   cl_kernel kernel;
   size_t global_work_size[2];
   size_t global_work_offset[2];

   /* CL input and output buffers */
   cl_mem bufferA;  // Input array on the device
   cl_mem bufferB;  // Output array on the device

   cl_uint i, j;

   /* read command line arguments */
   if (argc != 4)
   {
      printf("usage: %s n iterations PLATFORM_NUMBER\n", argv[0]);
      return 0;
   }
   n = atoi(argv[1]);                   /* number of rows (and columns) */
   iterations = atoi(argv[2]);          /* determines when to stop */
   int PLATFORM_NUMBER = atoi(argv[3]); /* which OpenCL platform to use, 0 or 1 */
   if ( !(PLATFORM_NUMBER==0 || PLATFORM_NUMBER==1) )
   {
      printf("Please choose PLATFORN_NUMBER 0 or 1 (not %d)\n", PLATFORM_NUMBER);
      return 0;
   }


   /* Host data buffers */
   InitializeData(&new, &val, n);
   //WriteData("data2.raw", val);

   // Compute the size of the OpenCL data buffers
   size_t datasize = sizeof(float) * (n * n);

   // Copy the data from an "array of pointers to rows" into
   // an "array of rows" that is compatible with OpenCL.
   float* clVal = (float*)malloc( datasize );
   for (i = 0; i < n; i++)
      for (j = 0; j < n; j++)
         clVal[i*n+j] = val[i][j];


   // Use this to check the output of each API call
   cl_int err;


   //-----------------------------------------------------
   // STEP 1: Discover and initialize a platform
   //-----------------------------------------------------

   // Use clGetPlatformIDs() to fill in the platform data structures.
   err = clGetPlatformIDs(0, NULL, &i);            // get number of platforms
   printf("The number of platforms is %d.\n", i);
   err = clGetPlatformIDs(2, platform, NULL);      // get 2 platform data strunctures

   if(CL_SUCCESS != err)
   {  printf("Couldn't find any platforms: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   //-----------------------------------------------------
   // STEP 2: Discover and initialize a GPU or CPU device
   //-----------------------------------------------------

   // Change PLATFORM_NUMBER if you want to use the other platform.
// PLATFORM_NUMBER = 0;

   // Use clGetDeviceIDs() to fill in a device data structure.
   err = clGetDeviceIDs(platform[PLATFORM_NUMBER], CL_DEVICE_TYPE_ALL, 1, NULL, &i);
   printf("The number of devices on platform %d is %d.\n", PLATFORM_NUMBER, i);
   // use the first device on the choosen platfrom
   err = clGetDeviceIDs(platform[PLATFORM_NUMBER], CL_DEVICE_TYPE_ALL, 1, &device, NULL);

   if(CL_SUCCESS != err)
   {  printf("Couldn't find any devices: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   //-----------------------------------------------------
   // STEP 3: Create a context
   //-----------------------------------------------------

   // Use clCreateContext() to create a context data structure
   // and associate it with the GPU device.
   context = clCreateContext(NULL, 1, &device, NULL, NULL, &err);

   if(CL_SUCCESS != err)
   {  printf("Couldn't create a context: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   //-----------------------------------------------------
   // STEP 4: Create a command queue for the device
   //-----------------------------------------------------

   // Use clCreateCommandQueue() to create a command queue
   // and associate it with the GPU device.
   cmdQueue = clCreateCommandQueue(context, device, 0, &err);

   if(CL_SUCCESS != err)
   {  printf("Couldn't create the command queue: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   //-----------------------------------------------------
   // STEP 5: Create and compile the program
   //-----------------------------------------------------

   // Read program file and place content into a string buffer
   program_handle = fopen(PROGRAM_FILE, "r");
   if(NULL == program_handle)
   {  printf("Couldn't find the kernel program file: %s", PROGRAM_FILE);
      exit(1);
   }
   fseek(program_handle, 0, SEEK_END);
   program_size = ftell(program_handle);
   rewind(program_handle);
   program_buffer = (char*)malloc(program_size + 1);
   if (NULL == program_buffer)
   {  perror("malloc");
      exit(-1);
   }
   program_buffer[program_size] = '\0';
   fread(program_buffer, sizeof(char), program_size, program_handle);
   fclose(program_handle);

   // Use clCreateProgramWithSource() to create a program data structure.
   program = clCreateProgramWithSource(
       context,
       1,
       (const char**)&program_buffer,
       &program_size,
       &err);

   free(program_buffer);

   if(CL_SUCCESS != err)
   {  printf("Couldn't create the program: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   // Use clBuildProgram() to build (compile) a program for the device.
   err = clBuildProgram(program,
                              0,
                           NULL,
                           "",     // compiler options
                           NULL,
                           NULL);

   if(CL_SUCCESS != err)
   {  /* Find size of log and print to std output */
      size_t log_size;
      clGetProgramBuildInfo(program, device, CL_PROGRAM_BUILD_LOG,
            0, NULL, &log_size);
      char *program_log = (char*) malloc(log_size + 1);
      program_log[log_size] = '\0';
      clGetProgramBuildInfo(program, device, CL_PROGRAM_BUILD_LOG,
            log_size + 1, program_log, NULL);
      printf("Could not compile the kernel program: ");
      sclPrintErrorFlag( err );
      printf("Build Log: ");
      printf("%s\n", program_log);
      free(program_log);
      exit(1);
   }

   //-----------------------------------------------------
   // STEP 6: Create the kernel
   //-----------------------------------------------------

   // Use clCreateKernel() to create a kernel from the 2d_sor function.
   kernel = clCreateKernel(program, KERNEL_FUNC, &err);

   if(CL_SUCCESS != err)
   {  printf("Couldn't create the kernel: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   //-----------------------------------------------------
   // STEP 7: Create device buffers to hold input and output data.
   //-----------------------------------------------------

   // Use clCreateBuffer() to create a CL buffer object
   // that will contain the data from the host array clVal.
   bufferA = clCreateBuffer(
       context,
       CL_MEM_READ_ONLY,// | CL_MEM_COPY_HOST_PTR,
       datasize,
       NULL,            // clVal,
       &err);

   if(CL_SUCCESS != err)
   {  printf("Couldn't create a buffer object: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   // Use clCreateBuffer() to create a CL buffer object
   // with enough space to hold the output data.
   bufferB = clCreateBuffer(
       context,
       CL_MEM_WRITE_ONLY,
       datasize,
       NULL,
       &err);

   if(CL_SUCCESS != err)
   {  printf("Couldn't create a buffer object: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   //-----------------------------------------------------
   // STEP 8: Write host data to device buffers
   //-----------------------------------------------------

   // Use clEnqueueWriteBuffer() to write data array "clVal"
   // to the device input buffer, bufferA.
   err = clEnqueueWriteBuffer(
       cmdQueue,
       bufferA,
       CL_FALSE,
       0,
       datasize,
       clVal,
       0,
       NULL,
       NULL);

   if(CL_SUCCESS != err)
   {  printf("Couldn't enqueue the write buffer command: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   // Use clEnqueueWriteBuffer() to write data array "clVal"
   // to the device output buffer, bufferB (this makes sure that
   // the boundary data ends up in the output buffer).
   err = clEnqueueWriteBuffer(
       cmdQueue,
       bufferB,
       CL_FALSE,
       0,
       datasize,
       clVal,
       0,
       NULL,
       NULL);

   if(CL_SUCCESS != err)
   {  printf("Couldn't enqueue the write buffer command: ");
      sclPrintErrorFlag( err );
      exit(1);
   }


   // Run the kernel the appropriate number of times.
   // For each iteration of the kernel we reverse the roles
   // of our two buffers. So each buffer alternates between
   // being the input buffer and the output buffer.
   for(int k = 0; k < iterations; k++)
   {
      //-----------------------------------------------------
      // STEP 9: Create kernel arguments from the CL buffers
      //-----------------------------------------------------

      // Use clSetKernelArg() to associate input and output buffers with the kernel
      err  = clSetKernelArg(
          kernel,
          0,
          sizeof(cl_mem),
          &n);

      err |= clSetKernelArg(
          kernel,
          (k%2==0)?1:2,    // switch the order of the buffers with each iteration
          sizeof(cl_mem),
          &bufferA);

      err |= clSetKernelArg(
          kernel,
          (k%2==0)?2:1,   // switch the order of the buffers with each iteration
          sizeof(cl_mem),
          &bufferB);

      if(CL_SUCCESS != err)
      {  printf("Couldn't set the kernel arguments: ");
         sclPrintErrorFlag( err );
         exit(1);
      }

      //-----------------------------------------------------
      // STEP 10: Enqueue the kernel for execution
      //-----------------------------------------------------

      // The work items are in a 2-dimensional work space and
      // there are (n-2) "work-items" in eacn dimension.
      global_work_size[0] = n-2;
      global_work_size[1] = n-2;
      // We need a global ID offset of 1 in each dimension
      // of the work space.
      global_work_offset[0] = 1;
      global_work_offset[1] = 1;

      // Use clEnqueueNDRangeKernel() to execute the kernel.
      err = clEnqueueNDRangeKernel(
          cmdQueue,
          kernel,
          2,        // 2-diensional work space
          global_work_offset,
          global_work_size,
          NULL,
          0,
          NULL,
          NULL);

      if(CL_SUCCESS != err)
      {  printf("Couldn't enqueue the kernel execution command: ");
         sclPrintErrorFlag( err );
         exit(1);
      }

   }// iterate the kernel again


   //-----------------------------------------------------
   // STEP 11: Read the output data back to the host
   //-----------------------------------------------------

   // Use clEnqueueReadBuffer() to read the OpenCL output
   // buffer (bufferC) to the host output array.
   clEnqueueReadBuffer(
       cmdQueue,
       bufferB,
       CL_TRUE,
       0,
       datasize,
       clVal,       // use the "clVal" array for the output
       0,
       NULL,
       NULL);

   if(CL_SUCCESS != err)
   {  printf("Couldn't enqueue the kernel execution command: ");
      sclPrintErrorFlag( err );
      exit(1);
   }

   //-----------------------------------------------------
   // STEP 12: Release OpenCL resources
   //-----------------------------------------------------

   clReleaseMemObject(bufferA);
   clReleaseMemObject(bufferB);
   clReleaseKernel(kernel);
   clReleaseProgram(program);
   clReleaseCommandQueue(cmdQueue);
   clReleaseContext(context);

   // Copy the data from an "array of rows"
   // back into an "array of pointers to rows".
   for (i = 0; i < n; i++)
      for (j = 0; j < n; j++)
         val[i][j] = clVal[i*n+j];

   /* Write the data to a file so that it can be visualized */
   WriteData("data.raw", val);
/*
      // for debugging purposes, print the data buffer
      for (i = 0; i < n; i++)
      {  for (j = 0; j < n; j++)
            printf( "%f ", clVal[i*n+j] );
         printf("\n");
      }
      printf("\n");
*/
   return 0;
}/* main() */
